/********************************************************************************/
/*    H   H Y   Y TTTTT EEEEE  CCC       HYTEC ELECTRONICS LTD                  */
/*    H   H  Y Y    T   E     C          5 Cradock Road,                        */
/*    HHHHH   Y     T   EEE   C          Reading, Berks.    Tel: 0118 9757770	*/
/*    H   H   Y     T   E     C          RG2 0JT            Fax: 0118 9757566	*/
/*    H   H   Y     T   EEEEE  CCC       Web: www.hytec-electronics.co.uk       */
/********************************************************************************/
/********************************************************************************/
/*      ________________________________________________________________        */
/*      | H Y T E C   8 4 1 7  2 4 B i t   A D C  A s y n  D r i v e r |        */
/*           -----------------------------------------------------------        */
/*                                                                              */
/*  Source file name      :-      drvHy8413ip.c                                 */
/*                                                                              */
/*  Initial creation date :-      15-Apr-2011                                   */
/*                                                                              */
/*  Original Developers   :-      Jim Chen						                */
/*                                                                              */
/********************************************************************************/
/*                                                                              */
/*  Description :- This code is for Hytec 8413 ADC EPICS support. It is based   */
/*  on Hy8417ipasyn driver code. It is osi complaint.		                    */
/*                                                                              */
/*  (C)2011 Hytec Electronics Ltd.                                              */
/*                                                                              */
/********************************************************************************/
/*                                                                              */
/* Modification record :-                                                       */
/*                                                                              */
/*  Date modified |  Person   | Version | Description of modification           */
/* ---------------|-----------|---------|-------------------------------------- */
/*    15/04/11    |  JC		  | 2.3     | Initial creation with new asyn model  */    
/*    09/11/11    |  JC		  | 2.4     | Transferred to latest asyn model 		*/    
/*    24/05/12    |  JC		  | 2.5     | Added clearing rebooting flag at the	*/    
/*                |           |         | beginning to avoid self hanging for   */    
/*                |           |         | somw VxWorks system                   */    
/*    10/10/14    |  JC		  | 2.6     | Modified to support 64bit Linux based	*/    
/*                |  M.Rivers |         | on Mark Rivers' driverTable model     */    
/*                |           |         | Note: need V2.6 or later of 			*/    
/*                |           |         | drvHyLinuxDriver.c in ipac module     */    
/*                |           |         | to work with 64bit Linux		        */    
/*                |           |         |    		                            */    
/********************************************************************************/

#define DRIVER_VERSION	206

/******************/
/* Hytec Module includes */
/******************/
#include "drvHy8413Asyn.h"

static Hy8413Asyn *driverTable[MAX_CARDS];
static int numCards;

/********************************************************************/
/*														           */
/* Public Sector												  */
/*													             */
/****************************************************************/
/* Constructor:
 *
 * Where: -
 *     portName = asyn port name
 *   carrierNum = carrier index number when adding a carrier card by ipacAddHy8002 call
 *    ipSlotNum = IP slot number on the carrier board (0 = A, 1 = B, 2 = C, 3 = D, 4 = E, 5 = F),
 *    vectorNum = interrupt vector (0 = find one for me),
 *         mode = 0 = continuous, 1 = triggered, 3 = gated,
 *      range   = 0 = -10V ~ +10V, 1 = -5V ~ +5V 
 *		format  = 0 = 2's complement, 1 = straight binary
 *    clockRate = 0  = 1Hz,    1 = 2Hz,    2 = 5Hz,    3 = 10Hz,
 *				  4  = 20Hz,   5 = 50Hz,   6 = 100Hz,  7 = 200Hz,
 *				  8  = 500Hz,  9 = 1kHz,  10 = 2kHz,  11 = 5kHz,
 *				  12 = 10khz, 13 = 20Khz, 14 = 50kHz, 15 = 100kHz.
 *				  16 = 160kHz
 *     extClock = 0 = internal, 1 = extrenal,
 */
    
Hy8413Asyn::Hy8413Asyn(char *portName, int carrierNum, int ipSlotNum, int vectorNum, int mode, 
		int range, int format, int clockRate, int extClock)    
		: asynPortDriver(portName, MAX_ADC_CHANNELS, NUM_PARAMS, 
		asynInt32Mask | asynInt32ArrayMask | asynFloat64Mask | asynFloat64ArrayMask | asynDrvUserMask,
        asynInt32Mask | asynInt32ArrayMask | asynFloat64Mask | asynFloat64ArrayMask, 
        ASYN_MULTIDEVICE, 1, /* ASYN_CANBLOCK=0, ASYN_MULTIDEVICE=1, autoConnect=1 */
        0, 0)  /* Default priority and stack size */

{
    const char *hytecstr = "(Hytec Electronics Ltd., Reading, UK)";
    int strok;
  	char lstr[7];
    int ishytec;
    int ismodel;
	int ver, i, j;
	volatile USHORT *calpointer;
	USHORT idmem;
	
	/* normal data channels definition */
    createParam(hy8413DataString, 			asynParamInt32, 	&hy8413Data_);
    createParam(hy8413FloatDataString, 		asynParamFloat64, 	&hy8413FloatData_);

	/* controls definition */
    createParam(hy8413ResetFIFOString, 		asynParamInt32, 	&hy8413resetFIFO_);
    createParam(hy8413SetFormatString, 		asynParamInt32, 	&hy8413SetFormat_);
    createParam(hy8413SetModeString, 		asynParamInt32, 	&hy8413SetMode_);
    createParam(hy8413RangeString, 			asynParamInt32, 	&hy8413Range_);
    createParam(hy8413EnableString, 		asynParamInt32, 	&hy8413Enable_);
    createParam(hy8413SoftTriggerString, 	asynParamInt32, 	&hy8413SoftTrigger_);
    createParam(hy8413AutoReenableString, 	asynParamInt32, 	&hy8413AutoReenable_);
    createParam(hy8413SetClockRateString, 	asynParamInt32, 	&hy8413SetClockRate_);
    createParam(hy8413SetExtClockString, 	asynParamInt32, 	&hy8413SetExtClock_);


    createParam(hy8413CounterString,		asynParamInt32, 	&hy8413Counter_);
    createParam(hy8413TriggerStatusString,	asynParamInt32, 	&hy8413TriggerStatus_);
    createParam(hy8413FWVersionString, 		asynParamInt32, 	&hy8413FWVersion_);
    createParam(hy8413DriverVersionString,	asynParamInt32, 	&hy8413DriverVersion_);
    createParam(hy8413SupportString, 		asynParamInt32, 	&hy8413Support_);
	
	/* Save the parameters in the driver structure    */
    this->carrierNum		 = carrierNum;
    this->ipSlotNum			 = ipSlotNum;
    this->vectorNum          = vectorNum;
    this->range          	 = range;
    this->mode          	 = mode;
    this->format          	 = format;
    this->clockrate        	 = clockRate;
    this->extclock           = extClock;
	
	/* Map Register address spaces  */
	this->adc = (IO_MEMORY *) ipmBaseAddr(carrierNum, ipSlotNum, ipac_addrIO);
    if (this->adc == NULL)
    {  
        /* can't register card register address space */
    	printf("configHy8413: Unable to register IO space\n");
        return;
    }

    /* Compare the HY8413 prom header, manufacturer's ID and card model
     * number with the expected values. */
  	idmem =  *((USHORT*)this->adc->prom.vita4);
  	if((idmem == 0x5649) || (idmem == 0x4956));
  	{
  		lstr[0] = 0x56; lstr[1] = 0x49;
  	}
  	idmem =  *((USHORT*)this->adc->prom.vita4 + 1);
  	if((idmem == 0x5441) || (idmem == 0x4154));
  	{	
  		lstr[2] = 0x54; lstr[3] = 0x41;
  	}
  	idmem =  *((USHORT*)this->adc->prom.vita4 + 2);
  	if((idmem == 0x3420) || (idmem == 0x2034));
  	{	
  		lstr[4] = 0x20; lstr[5] = 0x34;
  	}
  	lstr[6] = 0;

    strok = memcmp(lstr, VITA4, sizeof(this->adc->prom.vita4)) == 0;
    ishytec = this->adc->prom.hytecIdHigh == HYTECID_HI  &&
                  this->adc->prom.hytecIdLow == HYTECID_LO;
    ismodel = this->adc->prom.model == PROM_MODEL;

	/* check if we have been asked to be more chatty */
    /* start with some basic information */
    printf("PROM: Contents at offset - %#08x\n", (int)&this->adc->prom);
    printf("PROM: Prom header string - '%6s'\n", lstr);
    printf("PROM: Manufacturer's ID  - 0x%04x%04x %s\n",
        (int)this->adc->prom.hytecIdHigh, (int)this->adc->prom.hytecIdLow, 
        ishytec ? hytecstr : "(Unknown)");
    printf("\n");
    /* output a bit more about this specific model and card */
    printf("PROM: Model #/revision   - %#04hx rev. %#04hx\n",
        (int)this->adc->prom.model, (int)this->adc->prom.revision);

	ver = (int)this->adc->prom.revision;
	ver = ver/4096*1000 + (ver%4096)/256*100 + ((ver%4096)%256)/16*10 + ((ver%4096)%256)%16;
	setIntegerParam(hy8413FWVersion_, ver);

    printf("PROM: Serial number      - %hu\n", (int)this->adc->prom.serialNumber);
    printf("PROM: Driver IDs         - %#04hx, %#04hx\n",
        (int)this->adc->prom.driverIdHigh, (int)this->adc->prom.driverIdLow);
    printf("PROM: Device flags       - %#04hx\n", (int)this->adc->prom.flags);
    printf("PROM: Bytes used         - %#04hx (%hd)\n",
        (int)this->adc->prom.byteCount, (int)this->adc->prom.byteCount);
    printf("\n");

    /* check if everything passed okay */
    if (!strok)
        printf("PROM: INVALID PROM HEADER; EXPECTED '%s'\n", VITA4);
    if (!ishytec)
        printf("PROM: UNSUPPORTED MANUFACTURER ID; EXPECTED 0x%04x%04x, %s\n",
            HYTECID_HI, HYTECID_LO, hytecstr);
    if (!ismodel)
	{
        printf("PROM: UNSUPPORTED MODEL NUMBER; EXPECTED %#04hx\n",
            PROM_MODEL);
	}

    /* fetch calibration data +/-10V */
	calpointer = (USHORT*) &this->adc->prom.vita4[0];
	for(j=0; j<3; j++)
	{
	    this->adc->reg.acr &= 0xFF8F;												/* clear PG2 ~ PG0 */
	    this->adc->reg.acr |= ((j+1)<<4);											/* Set to get +/-10V channel 0-5 */
		for(i=0; i<6; i++)
		{
			cal10V[i+j*6].ngFS = (USHORT)*(calpointer + 0 + i*5);
			cal10V[i+j*6].ngHS = (USHORT)*(calpointer + 1 + i*5);
			cal10V[i+j*6].zero = (USHORT)*(calpointer + 2 + i*5);
			cal10V[i+j*6].psHS = (USHORT)*(calpointer + 3 + i*5);
			cal10V[i+j*6].psFS = (USHORT)*(calpointer + 4 + i*5);
			if((j == 2) && (i == 3)) break;
		}
	}
    
    /* fetch calibration data for +/-5V */
	for(j=0; j<3; j++)
	{
	    this->adc->reg.acr &= 0xFF8F;												/* clear PG2 ~ PG0 */
	    this->adc->reg.acr |= ((j+4)<<4);											/* Set to get +/-10V channel 0-5 */
	    for(i=0; i<6; i++)
	    {
	    	cal5V[i+j*6].ngFS = (USHORT)*(calpointer + 0 + i*5);
	    	cal5V[i+j*6].ngHS = (USHORT)*(calpointer + 1 + i*5);
	    	cal5V[i+j*6].zero = (USHORT)*(calpointer + 2 + i*5);
	    	cal5V[i+j*6].psHS = (USHORT)*(calpointer + 3 + i*5);
	    	cal5V[i+j*6].psFS = (USHORT)*(calpointer + 4 + i*5);
			if((j == 2) && (i == 3)) break;
	    }
    }  
    this->adc->reg.acr &= 0xFF8F;													/* clear PG2 ~ PG0 to restore */
    
    /* set driver version */
	setIntegerParam(hy8413DriverVersion_, DRIVER_VERSION);
	
    /* clear asyn message sent counters */
    this->messagesSent = 0;
    this->messagesFailed = 0;

    /* set up interrupt, work queue and mutex stuff */
    this->adc->reg.vector = vectorNum;            /* writes vector register offset 0x08 */
    driverTable[numCards] = this;
    numCards++;
	if (ipmIntConnect(this->carrierNum, this->ipSlotNum, this->vectorNum, INTSRC, numCards-1) < 0) {
      printf("configHy8413: failure connecting to interrupt\n");
      return;
    }

    ipmIrqCmd(this->carrierNum, this->ipSlotNum, 0, ipac_irqEnable);

    this->intMsgQId = epicsMessageQueueCreate(MAX_MESSAGES, 
    					(MAX_ADC_CHANNELS + 2)*sizeof(int));
    if (epicsThreadCreate("Hy8413intQueuedTask",
    					epicsThreadPriorityLow,
                        epicsThreadGetStackSize(epicsThreadStackMedium),
                        (EPICSTHREADFUNC)intQueuedTaskC,
                        (void *)this) == NULL)
	{	
      	printf("configHy8413: failure creating message queue.\n");
      	return;
	}
    
	/*******************************************************************/
	/************* Working out real stuffs for ADC *********************/
	/*******************************************************************/

	this->rebooting = 0;

	setupADC();

	// tell listeners the changes
	callParamCallbacks();
   	epicsAtExit(rebootCallbackC, this);
}


/*********************************************************/
/*                                                      */
/* Asyn interface Function implementations             */
/*                                                    */
/*****************************************************/

/* asynCommon methods */
void Hy8413Asyn::report(FILE *fp, int details)
{
	int i, mode, range, format, extclock, clockrate, fwver, swver;

    fprintf(fp, "Port: %s, carrier slot %d slot %d, base address=%p\n", 
            this->portName, this->carrierNum, this->ipSlotNum, (void *)this->adc);
    if (details == 1) 
    {
		getIntegerParam(hy8413SetMode_, &mode);
		getIntegerParam(hy8413Range_, &range);
		getIntegerParam(hy8413SetFormat_, &format);
		getIntegerParam(hy8413SetClockRate_, &clockrate);
		getIntegerParam(hy8413SetExtClock_, &extclock);
		getIntegerParam(hy8413FWVersion_, &fwver);
		getIntegerParam(hy8413DriverVersion_, &swver);
		
        fprintf(fp, "  Current csr=%x\n", this->adc->reg.csr);
        fprintf(fp, "  INT Vector=%x\n", this->vectorNum);
        fprintf(fp, "  Mode=%s\n", mode==0?"Continuous":"Trigger");
        fprintf(fp, "  Range=%s\n", range==0?"+/-10V":"+/-5V");
        fprintf(fp, "  Format=%s\n", format==0?"2's complement":"Straight binary");
        fprintf(fp, "  Clock Source=%s\n", extclock==0?"Internal":"External");
        fprintf(fp, "  Clock Rate=%d\n", clockrate);
        fprintf(fp, "  Firmware Version=%d\n", fwver);
        fprintf(fp, "  Driver Version=%d\n", swver);
        fprintf(fp, "  messages sent OK=%d; send failed (queue full)=%d\n",
                this->messagesSent, this->messagesFailed);       
    }
    if (details == 2)
    {
    	for(i=0; i<MAX_ADC_CHANNELS; i++)
    	{
            fprintf(fp, "  +/-10V chan=%d ngFS=%x ngHS=%x zero=%x psHS=%x psFS=%x\n", i, 
            		this->cal10V[i].ngFS, this->cal10V[i].ngHS, this->cal10V[i].zero,
            		this->cal10V[i].psHS, this->cal10V[i].psFS);
    	}
    	fprintf(fp, "\n");
    	for(i=0; i<MAX_ADC_CHANNELS; i++)
    	{
            fprintf(fp, "  +/-5V chan=%d ngFS=%x ngHS=%x zero=%x psHS=%x psFS=%x\n", i, 
            		this->cal5V[i].ngFS, this->cal5V[i].ngHS, this->cal5V[i].zero,
            		this->cal5V[i].psHS, this->cal5V[i].psFS);
    	}
    }
    asynPortDriver::report(fp, details);
}


/*****         asynInt32 methods       ******/
asynStatus Hy8413Asyn::writeInt32(asynUser *pasynUser, epicsInt32 value)
{
	asynStatus status = asynSuccess;
	epicsInt32 channel;
	int function = pasynUser->reason;
	status = 
    this->getAddress(pasynUser, &channel);
	
	/* decide what to do depending on which command we received */
	if (function == hy8413SetMode_)
		status = setMode( value );
	else if (function == hy8413Enable_)
		status = setEnable( value );
	else if (function == hy8413AutoReenable_)
		status = setAutoReEnable( value );
	else if (function == hy8413SoftTrigger_)
		status = softwareTrigger();
	else if (function == hy8413SetClockRate_)
		status = setClockRate( value );
	else if (function == hy8413SetExtClock_)
		status = setExtClock( value );
	else if (function == hy8413Range_)
		status = setRange( value );
	else if (function == hy8413SetFormat_)
		status = setFormat( value );
	else if (function == hy8413resetFIFO_)
		status = resetFIFO();

	return( status );
}

asynStatus Hy8413Asyn::readInt32(asynUser *pasynUser, epicsInt32 *value)
{
	asynStatus status = asynSuccess;
	epicsInt32 channel;
	int function = pasynUser->reason;
	int mode, triggerState, ver;
    if (this->rebooting) epicsThreadSuspendSelf();
  	
	getIntegerParam(hy8413SetMode_, &mode);
	getIntegerParam(hy8413TriggerStatus_, &triggerState);

	this->getAddress(pasynUser, &channel);

	if(function == hy8413Data_)
	{
		*value = (epicsInt32) calibration(this->adc->reg.data[channel], channel);
/* Testing */
/*if(channel == 8)
{
ver = this->adc->reg.csr;
mode = this->adc->reg.acr;
printf("chn=%d v=%x cv=%x acr=%x\n", channel, this->adc->reg.data[channel], *value, mode);
}*/
	}
	else if(function == hy8413Support_)
		*value = DRIVER_SUPPORT;

	else if(function == hy8413FWVersion_)
	{
		getIntegerParam(hy8413FWVersion_, &ver);
        *value = ver;                                               	 				/* firmware version */
	}
	else if(function == hy8413DriverVersion_)
	{
		getIntegerParam(hy8413DriverVersion_, &ver);
        *value = ver;                                               	 				/* driver version */
	}
	else if(function == hy8413TriggerStatus_)
	{
        if (mode == hy8413Continuous) 
        	*value = TRIGGERRED_COLLECTING;
        else
        	*value = triggerState;
	}
	else if(function == hy8413Counter_)
		*value = (epicsInt32) this->adc->reg.extCounter;
	
	return( status );
}

asynStatus Hy8413Asyn::getBounds(asynUser *pasynUser, epicsInt32 *low, epicsInt32 *high)
{
	*low = 0;
	*high = 65535;
	/* not usually implemented... See asyn documentation for details if needed. */
	return( asynSuccess );
}

asynStatus Hy8413Asyn::readFloat64(asynUser *pasynUser, epicsFloat64 *value)
{
	asynStatus status = asynSuccess;
	epicsInt32 channel;
	int function = pasynUser->reason;

    if (this->rebooting) epicsThreadSuspendSelf();

	this->getAddress(pasynUser, &channel);
	
    if(function == hy8413FloatData_)
	{
		*value = toVoltage(calibration(this->adc->reg.data[channel], channel));        		 
    }
	return( status );
}

/*********************************************************/
/*                                                      */
/* Private Sector                                      */
/*                                                    */
/*****************************************************/

void Hy8413Asyn::setupADC()
{
	if (this->rebooting)  epicsThreadSuspendSelf();

	this->adc->reg.csr = CSR_RSTFIFO; 												/* reset FIFO and initiate csr */   
	this->adc->reg.csr &= ~CSR_RSTFIFO; 											/* reset FIFO */   
	this->adc->reg.acr = ACR_NS; 											 		/* normal mode */   
    
	setClockRate( this->clockrate );	          
	setExtClock( this->extclock );	              
	setFormat( this->format );
	setRange( this->range );
	setEnable(0);																	/* clear enable */
	setAutoReEnable(0);																/* clear auto reenable */
	setMode( this->mode );
}

asynStatus Hy8413Asyn::setClockRate( int rate  )
{
	int mode, extClock;
	if (this->rebooting)  epicsThreadSuspendSelf();
	getIntegerParam(hy8413SetExtClock_, &extClock);
	getIntegerParam(hy8413SetMode_, &mode);

	/* clock rate only applys to internal clock */
	if(extClock == 1) return(asynError);
	
    if ((rate < 0) || (rate > 16)) return( asynError );

    this->adc->reg.clockRate = rate;
	setIntegerParam(hy8413SetClockRate_, rate);
    if(mode == hy8413Trigger) resetFIFO();											//if trigger mode, reset FIFO
	callParamCallbacks();
	return(asynSuccess);
}

asynStatus Hy8413Asyn::setExtClock( int ext )
{
	int mode;
	if (this->rebooting)  epicsThreadSuspendSelf();
	getIntegerParam(hy8413SetMode_, &mode);
       
    if( ext )                                                   					/* use external clock */
	   this->adc->reg.csr |= CSR_XC;
	else
	   this->adc->reg.csr &= ~CSR_XC;
    
    if(mode == hy8413Trigger) resetFIFO();											//if trigger mode, reset FIFO
	setIntegerParam(hy8413SetExtClock_, ext);
	callParamCallbacks();
    return( asynSuccess );
}

asynStatus Hy8413Asyn::softwareTrigger()
{
	int mode, enable;
	if (this->rebooting)  epicsThreadSuspendSelf();

	getIntegerParam(hy8413SetMode_, &mode);
	getIntegerParam(hy8413Enable_, &enable);
 
    if(mode != hy8413Trigger ) 
        return asynError;                                           				/* software trigger at realtime only applies to trigger mode */
    if(enable)                                           							/* only trigger when enabled */
    {
        this->adc->reg.csr &= ~CSR_ST;                           					/* clear software trigger latch */
        this->adc->reg.csr |= CSR_ST;                            					/* software trigger to kick off the acquisition */
        this->adc->reg.csr &= ~CSR_ST;                           					/* clear software latch to enable hardware trigger afterwards */
		setIntegerParam(hy8413TriggerStatus_, TRIGGERRED_COLLECTING);
  		callParamCallbacks();
    }
    return( asynSuccess );
}

asynStatus Hy8413Asyn::setMode( int mode )
{
	int enable;
	if (this->rebooting)  epicsThreadSuspendSelf();

	getIntegerParam(hy8413Enable_, &enable);
    
    if (mode == hy8413Continuous)
        this->adc->reg.csr &= ~CSR_ET;                    							/* disable trigger */
    else
        this->adc->reg.csr |= CSR_ET;                    							/* enable trigger */
    	
    setEnable(enable);																/* set enable as it was */
	setIntegerParam(hy8413SetMode_, mode);
	callParamCallbacks();
    return( asynSuccess );
}

asynStatus Hy8413Asyn::setEnable( int enable )
{
	int mode;
	if (this->rebooting)  epicsThreadSuspendSelf();

	getIntegerParam(hy8413SetMode_, &mode);
 
	if (enable)                                                     				/* if enable */
    {
        if (mode == hy8413Trigger)
        {
                this->adc->reg.csr |= CSR_ETF;                    					/* enable interrupt on external full */
                setIntegerParam(hy8413TriggerStatus_, WAITING_TRIGGER);
        }
        this->adc->reg.csr |= CSR_ARM;	                							/* Arm ADC, enable acquiring */
    }
	else                                                            				/* if disable */
    {
        if (mode == hy8413Trigger)
        {
                this->adc->reg.csr &= ~CSR_ETF;                  					/* disable trigger */
                setIntegerParam(hy8413TriggerStatus_, FINISHED_AVAILABLE);
        }
        this->adc->reg.csr &= ~CSR_ARM;	                							/* disarm ADC */
    }
	
	setIntegerParam(hy8413Enable_, enable);
	callParamCallbacks();
	return( asynSuccess );
}

asynStatus Hy8413Asyn::setAutoReEnable(int reenable)
{
	if (this->rebooting)  epicsThreadSuspendSelf();
	setIntegerParam(hy8413AutoReenable_, reenable);
	callParamCallbacks();
	return( asynSuccess );
}

asynStatus Hy8413Asyn::setRange(int range)
{
	int mode;
	if (this->rebooting)  epicsThreadSuspendSelf();
	getIntegerParam(hy8413SetMode_, &mode);

	if (this->rebooting)  epicsThreadSuspendSelf();
	if(range)
        this->adc->reg.acr |= ACR_RGE;	                							/* range = 1 to set +/-5V */
	else
        this->adc->reg.acr &= ~ACR_RGE;	                							/* range = 0 to set +/-10V */
		
	setIntegerParam(hy8413Range_, range);
    if(mode == hy8413Trigger) resetFIFO();											//if trigger mode, reset FIFO
	callParamCallbacks();
	return( asynSuccess );
}

asynStatus Hy8413Asyn::setFormat(int format)
{
	if (this->rebooting)  epicsThreadSuspendSelf();
	if(format)
        this->adc->reg.acr |= ACR_2C;	                							/* straight binary */
	else
        this->adc->reg.acr &= ~ACR_2C;	                							/* 2's complement */
	setIntegerParam(hy8413SetFormat_, format);
	callParamCallbacks();
	return( asynSuccess );
}

asynStatus Hy8413Asyn::resetFIFO()
{
	if (this->rebooting)  epicsThreadSuspendSelf();
    this->adc->reg.csr |= CSR_RSTFIFO;                  							/* reset FIFO */		
    this->adc->reg.csr &= ~CSR_RSTFIFO;                  							/* clear flag */		
	return( asynSuccess );
}

void Hy8413Asyn::rebootCallback()
{
	setEnable(0);                              										/* disable ADC by clearing the ARM bit  */
    this->rebooting = 1;				                            				/* flag we are rebooting */
}

/*********************************************************/
/*                                                      */
/* Floating Voltage Convertion Routines                */
/*                                                    */
/*****************************************************/

/* this routine convert the raw adc value to voltage, -10V ~ +10V or -5V ~ +5V as per the range setting.  */
double Hy8413Asyn::toVoltage( USHORT rawval )
{
	int range, format;
	double v;
	
	getIntegerParam(hy8413Range_, &range);
	getIntegerParam(hy8413SetFormat_, &format);

	v = range == 0 ? 10.0 : 5.0;												// decide range
	if(format)
		return (double)(v * ((double) 2 * rawval - 0xFFFF - 0x00) / ((double) 0xFFFF - 0x00));
	else
		return (double)(v * ((double) 2 * ((short)rawval) - 0x7FFF - (short)0x8000) / ((double) 0x7FFF - (short)0x8000));		
}

/* this routine convert the raw adc value to voltage, -10V ~ +10V or -5V ~ +5V as per the range setting.  */
USHORT Hy8413Asyn::calibration( USHORT rawval, int channel )
{
	int range, format;
	double rv=0, raw=rawval, ngFS, ngHS, zero, psHS, psFS;
	USHORT rtvalue;
	
	getIntegerParam(hy8413Range_, &range);
	getIntegerParam(hy8413SetFormat_, &format);

	/* if it is 2's complement, convert it to straight first */
	if(format == 0) 
	{
		if((rawval & 0x8000) == 0)
			raw = (double)rawval + 0x8000;
		else
		{
			rawval = ~rawval + 1;
			raw = (double)((USHORT)0x8000 -rawval);
		}
	}

	if(range == 0)																	/* +/-10V */
	{
		ngFS = (double)cal10V[channel].ngFS;
		ngHS = (double)cal10V[channel].ngHS;
		zero = (double)cal10V[channel].zero;
		psHS = (double)cal10V[channel].psHS;
		psFS = (double)cal10V[channel].psFS;
	}else
	{
		ngFS = (double)cal5V[channel].ngFS;
		ngHS = (double)cal5V[channel].ngHS;
		zero = (double)cal5V[channel].zero;
		psHS = (double)cal5V[channel].psHS;
		psFS = (double)cal5V[channel].psFS;		
	}
	if(raw > psHS)
		rv = ((double)raw - psHS)*0x3FF8/(psFS - psHS)+0xBFF8; 
	if((raw > zero) && (raw <= psHS))
		rv = ((double)raw - zero)*0x3FF8/(psHS - zero)+0x8000; 
	if((raw <= zero) && (raw >= ngHS))
		rv = ((double)raw - zero)*0x3FF9/(zero - ngHS)+0x8000; 
	if(raw < ngHS) 
		rv = ((double)raw - ngHS)*0x3FF9/(ngHS - ngFS)+0x4007; 

	if(rv >= 0xFFFF)
		rtvalue = 0xFFFF;
	else
		rtvalue = (USHORT)(rv);
	
	/* if it is 2's complement, convert it back to 2's cpmplement */
	if(format == 0) 
	{
		if(rtvalue > 0x8000)
			rtvalue = rtvalue - 0x8000;
		else
			rtvalue = ~((USHORT)0x8000 - rtvalue) - 1;
	}
	return rtvalue;
}



/*********************************************************/
/*                                                      */
/* External Wrapped Routines      					   */
/*                                                    */
/*****************************************************/

// These functions must have C linkage because they are called from other EPICS components
extern "C" {
static void rebootCallbackC(void *drvPvt)
{
    Hy8413Asyn *pPvt = (Hy8413Asyn *)drvPvt;
    pPvt->rebootCallback();
}

static void INTSRC(int cardNumber)
{
    Hy8413Asyn *pPvt = driverTable[cardNumber];
    pPvt->intSR();
}

static void intQueuedTaskC(void *drvPvt)
{
    Hy8413Asyn *pPvt = (Hy8413Asyn *)drvPvt;
    pPvt->intQueuedTask();	
}

}

/*********************************************************/
/*                                                      */
/* Interrupt and Queue Task and external Routines      */
/*                                                    */
/*****************************************************/
void Hy8413Asyn::intSR()
{
    int data[2];

    this->adc->reg.csr &= ((~CSR_ETF) & (~CSR_ARM));                  	/* disable interrupt and remove ARM bit */

	if (this->adc->reg.csr & CSR_TF)     											/* external FIFO full */
    {
        /* Wake up queue task which calls callback routines */
         if (epicsMessageQueueTrySend(this->intMsgQId, data, sizeof(data)) == 0)
         	messagesSent++;
         else
         	messagesFailed++;
    }
//    callParamCallbacks();
}

void Hy8413Asyn::intQueuedTask()
{
    int chan, pointer;
    int data[2];
	int enable, reenable;

    while(1) {
        /* Wait for event from interrupt routine */
        epicsMessageQueueReceive(intMsgQId, data, sizeof(data));
                 
     	getIntegerParam(hy8413AutoReenable_, &reenable);
     	getIntegerParam(hy8413Enable_, &enable);
  
        for (pointer=0; pointer<BANKSIZE; pointer++)
        {
            for (chan=0; chan<MAX_ADC_CHANNELS; chan++)
            {
            	rawData[chan][pointer] = calibration(this->adc->reg.extFIFO, chan);
            	adcData[chan][pointer] = toVoltage(rawData[chan][pointer]);
            }
        }

        /****************************************/
        /****************************************/
        /********* waveform collection **********/
        /****************************************/
        /****************************************/

        for(chan=0; chan<MAX_ADC_CHANNELS; chan++)
        {
			doCallbacksInt32Array((epicsInt32 *) &rawData[chan],
					BANKSIZE, hy8413Data_, chan);

 			doCallbacksFloat64Array((epicsFloat64 *) &adcData[chan],
 					BANKSIZE, hy8413FloatData_, chan);
		}

        if (reenable && enable)                                           				/* if re-trigger is on and enable is on */
            setEnable( ENABLEADC );                       								/* enable trigger */
        ipmIrqCmd(carrierNum, ipSlotNum, 0, ipac_irqEnable);
    }
    callParamCallbacks();
}

